/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**               Keyboard.c                                **/
/**                                                         **/
/** This file contains the procedure to handle keyboard     **/
/** events sent by the X-server. Also included are          **/
/** functions that are called on FocusIn and -Out events.   **/
/** the emulation and Doption procedure for handling        **/
/** changes to options.                                     **/
/** To be used together with the improved X11-drivers       **/ 
/** written by Arnold Metselaar.                            **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-2001                 **/
/**               Arnold Metselaar 2001                     **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify us, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include "MSX.h"
#include "Sound.h"
#include "Keyboard.h"
/** used after a FocusInEvent to process an XKeyMapEvent */
void OnKeyMap(XKeymapEvent *E);

int KeyBuf;
static byte XKeyMap[18];
extern int CurDisk[2];
extern char *Disks[2][MAXDISKS+1];
extern Display * Dsp;
extern Window Wnd;
extern int SndVolume, SndSwitch;
extern int FastForward;

/** Options specific to this keyboard driver */
static char * KeyOptions[]= {
  "keybuf", "nokeybuf", 0
} ;

/** process the driver specific options */
int 
KeyboardOption(int argc, char *argv[], int *p)
{
  int J;
  
  for(J=0 ; KeyOptions[J] ; J++)
    if(!strcmp(argv[(*p)]+1,KeyOptions[J])) break;
  switch (J) 
    {
    case 0: KeyBuf=1; break;
    case 1: KeyBuf=0; break;
    }
  return (KeyOptions[J]!=NULL);
}

/** Commands specific to this keyboard driver */
static char * KeyCommands[]= {
  "keybuf", "nokeybuf", 0
} ;

/** process the driver specific commands */
int 
KeyboardCommand(int argc, char *argv[], int *p)
{
  int J;
  
  for(J=0 ; KeyCommands[J] ; J++)
    if(!strcmp(argv[(*p)],KeyCommands[J])) break;
  switch (J) 
    {
    case 0: KeyBuf=1; break;
    case 1: KeyBuf=0; break;
    }
  return (KeyCommands[J]!=NULL);
}

/** called just after the window is mapped */
int 
InitKeyboard(Display *Dsp, Window Wnd)
{
  (void)Dsp;
  (void)Wnd; /* do nothing for now */
  memset(XKeyMap, 0xff, sizeof(XKeyMap));
  return 1;
}

/** called just after the colors are deallocated, before the rest of
    the shutdown */
void 
TrashKeyboard(Display *Dsp, Window Wnd)
{
  (void)Dsp;
  (void)Wnd; /* do nothing for now */
}

/** called when fMSX gains focus */
void 
OnFocusIn(XFocusChangeEvent * E)
{
  XEvent KMS;

  XAutoRepeatOff(E->display);
  XCheckWindowEvent(Dsp,0x0,KeymapStateMask,&KMS);
  OnKeyMap((XKeymapEvent *)(&KMS));
}

/** called when fMSX loses focus */
void 
OnFocusOut(XFocusChangeEvent * E)
{
  XAutoRepeatOn(E->display);
}

/*************************************************************/
/** Now comes the part that is mostly copied from the main  **/
/** distribution                                            **/
/*************************************************************/

/** Keyboard bindings ****************************************/
struct { word Code; byte Pos,Mask; } Keys[256] =
{
  {' ',8,0x01},{'\'',2,0x01},
  {'[',1,0x20},{'\\',1,0x10},{']',1,0x40},
  {',',2,0x04},{'-',1,0x04},{'.',2,0x08},{'/',2,0x10},
  {'0',0,0x01},{'1',0,0x02},{'2',0,0x04},{'3',0,0x08},
  {'4',0,0x10},{'5',0,0x20},{'6',0,0x40},{'7',0,0x80},
  {'8',1,0x01},{'9',1,0x02},{';',1,0x80},{'=',1,0x08},
  {'`',2,0x02},{'a',2,0x40},{'b',2,0x80},{'c',3,0x01},
  {'d',3,0x02},{'e',3,0x04},{'f',3,0x08},{'g',3,0x10},
  {'h',3,0x20},{'i',3,0x40},{'j',3,0x80},{'k',4,0x01},
  {'l',4,0x02},{'m',4,0x04},{'n',4,0x08},{'o',4,0x10},
  {'p',4,0x20},{'q',4,0x40},{'r',4,0x80},{'s',5,0x01},
  {'t',5,0x02},{'u',5,0x04},{'v',5,0x08},{'w',5,0x10},
  {'x',5,0x20},{'y',5,0x40},{'z',5,0x80},
  {XK_BackSpace,  7,0x20},
  {XK_Tab,        7,0x08},
  {XK_Return,     7,0x80},
  {XK_Escape,     7,0x04},
  {XK_Left,       8,0x10},
  {XK_Up,         8,0x20},
  {XK_Right,      8,0x80},
  {XK_Down,       8,0x40},
  {XK_F1,         6,0x20},
  {XK_F2,         6,0x40},
  {XK_F3,         6,0x80},
  {XK_F4,         7,0x01},
  {XK_F5,         7,0x02},
  {XK_Shift_L,    6,0x01}, /* SHIFT      */
  {XK_Shift_R,    6,0x01}, /* SHIFT      */
  {XK_Control_L,  6,0x02}, /* CONTROL    */
  {XK_Control_R,  6,0x08}, /* CAPSLOCK   */
  {XK_Alt_L,      6,0x04}, /* GRAPH      */
  {XK_Alt_R,      6,0x10}, /* CODE       */
  {XK_Delete,     8,0x08}, /* DELETE     */
  {XK_Insert,     8,0x04}, /* INSERT     */
  {XK_Break,      7,0x10}, /* STOP/BREAK */
  {XK_Home,       8,0x02}, /* HOME/CLS   */
  {XK_End,        7,0x40}, /* SELECT     */
  /* use row 16/17 and keypad for joystick */
  {XK_KP_Up,     16,0x01},
  {XK_KP_8,      16,0x01},
  {XK_KP_Down,   16,0x02},
  {XK_KP_2,      16,0x02},
  {XK_KP_Left,   16,0x04},
  {XK_KP_4,      16,0x04},
  {XK_KP_Right,  16,0x08},
  {XK_KP_6,      16,0x08},
  {XK_KP_Insert, 16,0x10},
  {XK_KP_0,      16,0x10},
  {XK_KP_Delete, 16,0x20},
  {XK_KP_Decimal,16,0x20},
 
  {0,0,0x00}               /** The End. **/
};

/** Called for each XKeyEvent received, should return nozero if no
    further keyevents are requested for this interrupt (key buffering).
    mostly copied from the main distribution */
int 
OnKeyEvent(XKeyEvent *E)
{
  int I,J;

  J=XLookupKeysym(E,0);
  if(E->type==KeyPress)
    {
      /* Special keys pressed... */
      switch(J)
	{
#ifdef DEBUG
        case XK_F11:
          CPU.Trace=!CPU.Trace;
          break;
#endif

        case XK_F6:
          AutoFire=!AutoFire;
          break;
	  
        case XK_F7:
          if(StateName)
          { 
	    if (E->state & ControlMask) SaveState(StateName); 
	    else LoadState(StateName); 
	  }
          break;

        case XK_F8:
          I=(E->state & ControlMask)? 1:0;
          if(Disks[I][0])
          {
            CurDisk[I]=(CurDisk[I]+1)%MAXDISKS;
            if(!Disks[I][CurDisk[I]]) CurDisk[I]=0;
            ChangeDisk(I,Disks[I][CurDisk[I]]);
          }
          break;

        case XK_F9:
	  if (E->state & ControlMask) UseFont=UseFont? 0:FontBuf? 1:0;
	  else if(!FastForward) { FastForward=UPeriod;UPeriod=10; }
          break;

        case XK_F10:
#ifdef SOUND
          if (! (E->state & ControlMask) )
	    SetChannels(SndVolume,SndSwitch=~SndSwitch);
#endif
          if (E->state & ControlMask) MIDILogging(MIDI_TOGGLE);
          break;

        case XK_F12:
          ExitNow=1;
          break;
	
	  /* treat joystick as ordinary keypresses -- Arnold */
	}

      /* Key pressed: reset bit in KeyMap */
      for(I=0;Keys[I].Code;I++)
        if(Keys[I].Code==J) { XKeyMap[Keys[I].Pos]&=~Keys[I].Mask;break; }
    }
  else
    {
      if (XK_F9==J)
	if(FastForward) { UPeriod=FastForward;FastForward=0; }
      /* Key released: set bit in KeyMap */
      for(I=0;Keys[I].Code;I++)
        if(Keys[I].Code==J) { XKeyMap[Keys[I].Pos]|=Keys[I].Mask;break; }
    }

  /* Copy local keymap to the global one */
  memcpy(KeyMap,XKeyMap,sizeof(KeyMap));

  return KeyBuf && Keys[I].Code;
}

/** called when fMSX receives an XKeyMapEvent */
void OnKeyMap(XKeymapEvent *E)
{
  XKeyEvent kev;
  int i,j;
  char mask;
  word ksym;

  /* elease all keys */
  memset(XKeyMap, 0xff, sizeof(XKeyMap));

  /* set up fake keypressevent */
  kev.type=KeyPress;
  kev.serial=E->serial;
  kev.send_event=1;
  kev.display=E->display;
  kev.window=E->window;
  kev.root=E->window;
  kev.subwindow=E->window;
  kev.time=0;
  kev.x=kev.y=0;
  kev.x_root=kev.y_root=0;
  kev.state=0;
  kev.same_screen=1;

  /* proces the all keys on the X-keyboard */
  for (i=j=0 ; i<32 ; ++i) 
    for (mask=1 ; mask & 0xFF ; mask <<= 1, ++j) 
      if (E->key_vector[i] & mask) { /* pressed ? */
	int I;

	kev.keycode=j;
	ksym=XLookupKeysym(&kev, 0);
        for (I=0;Keys[I].Code;I++)
          if (Keys[I].Code==ksym) { 
	    XKeyMap[Keys[I].Pos]&=~Keys[I].Mask;
	    break; 
	  }
      }
  /* Copy local keymap to the global one */
  memcpy(KeyMap,XKeyMap,sizeof(KeyMap));
}

/** Joystick() ***********************************************/
/** Query position of a joystick connected to port N.       **/
/** Returns 0.0.F2.F1.R.L.D.U.                              **/
/*************************************************************/
byte Joystick(register byte N) { return(~XKeyMap[16+N]); }


/** Mouse() **************************************************/
/** Query coordinates of a mouse connected to port N.       **/
/** Returns F2.F1.Y.Y.Y.Y.Y.Y.Y.Y.X.X.X.X.X.X.X.X.          **/
/**                                                         **/
/** Changed to get relative motions  -- Arnold              **/ 
/*************************************************************/
int Mouse(register byte N)
{
  Window Dummy;
  int X,Y,J,Buttons;
  static int oldX, oldY, Mflag;

  /* Get mouse coordinates */
  if (XQueryPointer(Dsp,Wnd,&Dummy,&Dummy,&J,&J,&X,&Y,&Buttons))
  {
    Buttons=(Buttons&Button1Mask? 0x10000:0x00000)
           |(Buttons&Button2Mask? 0x20000:0x00000)
           |(Buttons&Button3Mask? 0x20000:0x00000);
    if (Mflag)
      {
	J=X; X-=oldX; oldX=J;
	J=Y; Y-=oldY; oldY=J;
      }
    else 
      {
	oldX=X;
	oldY=Y;
	Mflag=1;
      }
  }
  else X=Y=Buttons=0;

  return( Buttons|(Y<<8)|X );
}

